from . import aliases
import loady, os
from .. util import json

guess_name = loady.importer.guess_name
CACHE = os.path.expanduser('~/.bibliopixel/code_cache')


def data(name, use_json=True):
    if not name:
        return {}

    try:
        return json.loads(name)
    except:
        return loady.data.load(name, use_json)


def code(name, python_path=None):
    return name and loady.code.load_code(name, python_path)


def module(name, python_path=None):
    return name and loady.code.load_module(name, python_path)


def extender(path):
    parts = [os.getcwd()] + path.split(':')
    return parts and loady.sys_path.extender(':'.join(parts))



import copy
from . import alias_lists, construct, merge
from .. import layout

DEFAULT_DRIVERS = [construct.to_type('simpixel')]


def cleanup_layout(animation):
    # Try to fill in the layout if it's missing.
    datatype = animation['datatype']

    try:
        args = datatype.LAYOUT_ARGS
        layout_cl = datatype.LAYOUT_CLASS
    except:
        raise ValueError('Missing "layout" section')

    args = {k: animation[k] for k in args if k in animation}
    return dict(args, datatype=layout_cl)


def cleanup_aliases(desc):
    alias_lists.PROJECT_ALIASES = desc.pop('aliases', {})
    return desc


def cleanup_numbers(desc):
    numbers = desc.pop('numbers', 'python')
    if numbers != 'python':
        desc.setdefault('maker', {})['numpy_dtype'] = numbers

    return desc


def cleanup_animation(desc):
    if not desc['animation']:
        raise ValueError('Missing "animation" section')

    desc['animation'] = construct.to_type_constructor(
        desc['animation'], 'bibliopixel.animation')
    datatype = desc['animation'].get('datatype')
    if not datatype:
        raise ValueError('Missing "datatype" in "animation" section')

    from .. animation import sequence
    if not issubclass(datatype, sequence.Sequence):
        # Magic here to allow this to work for non-sequences.
        length = desc['animation'].pop('length', [])
        if length:
            desc['run']['seconds'] = length[0]

    desc = merge.merge(getattr(datatype, 'PROJECT', {}), desc)

    run = desc.pop('run')
    anim_run = desc['animation'].setdefault('run', {})
    if run:
        desc['animation']['run'] = dict(run, **anim_run)
    return desc


def cleanup_drivers(desc):
    driver = construct.to_type(desc.pop('driver', {}))
    drivers = [construct.to_type(d) for d in desc['drivers']]
    if driver:
        if drivers:
            drivers = [dict(driver, **d) for d in drivers]
        else:
            drivers = [driver]

    desc['drivers'] = drivers or DEFAULT_DRIVERS
    return desc


def cleanup_dimensions(desc):
    desc = copy.deepcopy(desc)
    dimensions = desc.pop('dimensions', None)
    if not dimensions:
        return desc

    if len(desc['drivers']) != 1:
        raise ValueError('Cannot use dimensions with more than one driver')

    try:
        d = [int(dimensions)]
    except:
        d = list(dimensions)

    ldesc = construct.to_type_constructor(desc.get('layout') or {})
    driver = desc['drivers'][0]

    if len(d) == 1:
        driver['num'] = d[0]
        ldesc.setdefault('datatype', layout.strip.Strip)

    elif len(d) == 2:
        driver['num'] = d[0] * d[1]
        ldesc.setdefault('datatype', layout.matrix.Matrix)
        ldesc.update(width=d[0], height=d[1])

    elif len(d) == 3:
        driver['num'] = d[0] * d[1] * d[2]
        ldesc.setdefault('datatype', layout.cube.Cube)
        ldesc.update(x=d[0], y=d[1], z=d[2])

    else:
        raise ValueError('Dimension %s > 3' % len(d))

    desc['layout'] = ldesc
    return desc


def cleanup(desc):
    desc = cleanup_aliases(desc)
    desc = cleanup_animation(desc)
    desc = cleanup_drivers(desc)
    desc = cleanup_dimensions(desc)  # Must come after cleanup_drivers
    desc = cleanup_numbers(desc)

    return desc



import json, os, sys, yaml


# Allow open to be patched for tests.
open = __builtins__['open']


def loads(s, filename=''):
    if filename.endswith('.yml'):
        return yaml.load(s)
    return json.loads(s)


def dump(data, file=sys.stdout, **kwds):
    """Dump JSON data to a file or file handle"""
    def dump(fp):
        return json.dump(data, fp, indent=4, sort_keys=True, **kwds)

    if not isinstance(file, str):
        return dump(file)

    if os.path.isabs(file):
        parent = os.path.dirname(file)
        if not os.path.exists(parent):
            os.makedirs(parent, exist_ok=True)

    with open(file, 'w') as fp:
        return dump(fp)


def load(filename):
    """
    Loads not only JSON files but also YAML files ending in .yml.
    """
    if isinstance(filename, str):
        fp = open(filename)
    else:
        fp = filename
        filename = getattr(fp, 'name', '')

    try:
        return loads(fp.read(), filename)

    except Exception as e:
        e.args = ('There was a error in the data file', filename) + e.args
        raise



import struct
from . import websocket
from .. driver_base import DriverBase
from ... util import log, server_cache


class SimPixel(DriverBase):
    CACHE = server_cache.ServerCache(
        constructor=websocket.Server,
        selectInterval=0.001)

    def __init__(self, num=1024, port=1337, pixel_positions=None, **kwds):
        """
        Args:
            num:  number of LEDs being visualizer.
            port:  the port on which the SimPixel server is running.
            pixel_positions:  the positions of the LEDs in 3-d space.
            **kwds:  keywords passed to DriverBase.
        """
        super().__init__(num, **kwds)
        self.port = port
        self.pixel_positions = self.server = self.thread = None

        if pixel_positions:
            self.set_pixel_positions(pixel_positions)

    def start(self):
        self.server = self.CACHE.get_server(self.port)
        if self.pixel_positions:
            self.server.update(positions=self.pixel_positions)

    def set_pixel_positions(self, pixel_positions):
        # Flatten list of led positions.
        pl = [c for p in pixel_positions for c in p]
        self.pixel_positions = bytearray(struct.pack('<%sh' % len(pl), *pl))
        if self.server:
            self.server.update(positions=self.pixel_positions)

    def cleanup(self):
        self.server.close()

    def _compute_packet(self):
        self._render()

    def _send_packet(self):
        self.server.update(pixels=self._buf)


# This is DEPRECATED.
DriverSimPixel = SimPixel



"""
Test the all_pixel.
"""

DESCRIPTION = """
Equivalent to

    bp --num=10 --loglevel=debug --animation=strip_test \
      --driver=serial --fail_on_exception --layout=strip  --ledtype=<argument>
"""

from .. drivers import ledtype
from . import run as _run

LEDTYPES = """
BiblioPixel currently understands the following types of LED strips:

""" + ', '.join(sorted(ledtype.LEDTYPE.__members__.keys()))

LEDTYPE_HELP = """\
The type of the LED strip that is connected to your AllPixel.
""" + LEDTYPES

NO_LED_ERROR = """No ledtype provided.
""" + LEDTYPES


ARGUMENTS = {
    'animation': 'strip_test',
    'driver': 'serial',
    'fail_on_exception': True,
    'layout': 'strip',
    'loglevel': 'debug',
    'num': 10,
}


def run(args):
    for key, value in ARGUMENTS.items():
        setattr(args, key, value)

    args.ledtype = args.ledtype or args.ledtype_arg
    if not args.ledtype:
        raise ValueError(NO_LED_ERROR)

    _run.run(args)


def set_parser(args):
    _run.set_parser(args)
    args.set_defaults(run=run)
    args.add_argument('ledtype_arg', help=LEDTYPE_HELP, default='')



class AttributeDict(object):
    """A dict that exposes its values as attributes."""

    def __init__(self, *args, **kwds):
        for k, v in dict(*args, **kwds).items():
            setattr(self, k, v)

    def __setattr__(self, k, v):
        if isinstance(v, dict):
            v = AttributeDict(**v)
        super().__setattr__(k, v)

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        return self.__dict__ != other.__dict__

    def __bool__(self):
        return bool(self.__dict__)



import getpass, platform, sys
from .. util import log
from . control_source import ExtractedSource

# See https://stackoverflow.com/questions/42603000
DARWIN_ROOT_WARNING = """
In MacOS, pynput must to be running as root in order to get keystrokes.

Try running your program like this:

    sudo %s <your commands here>
"""

INSTALL_ERROR = """
Please install the pynput library with

    $ pip install pynput

"""


def keyname(key):
    return getattr(key, 'name', None) or getattr(key, 'char')


class Keyboard(ExtractedSource):
    KEYS_BY_TYPE = {'press': ['type', 'key'], 'release': ['type', 'key']}
    NORMALIZERS = {'key': keyname}

    def _make_thread(self):
        try:
            import pynput
        except Exception as e:
            e.args = (INSTALL_ERROR,) + e.args
            raise

        if platform.platform().startswith('Darwin'):
            if getpass.getuser() != 'root':
                log.warning(DARWIN_ROOT_WARNING, sys.argv[0])

        return pynput.keyboard.Listener(
            lambda key: self.callback({'type': 'press', 'key': key}),
            lambda key: self.callback({'type': 'release', 'key': key}))


def test_keyboard():
    keyboard = Keyboard(callback=print)
    keyboard.start()
    keyboard.thread.join()


if __name__ == '__main__':
    test_keyboard()



import threading
from . import extractor


class ControlSource():
    """
    A ControlSource sends messages to a callback.

    Usage:

        source = MyControlSource(callback)
        source.start()

        # callback gets periodically called in another thread with messages.

        source.stop()
    """

    def __init__(self, callback):
        self._callback = callback

    def start(self):
        self.thread = self._make_thread()
        self.thread.start()

    def stop(self):
        self.thread.stop()

    def callback(self, msg):
        self._callback(msg)

    def _make_thread(self):
        return threading.Thread(target=self.loop, daemon=True)

    def loop(self):
        for msg in self:
            self.callback(msg)

    def __iter__(self):
        raise NotImplementedError


class ExtractedSource(ControlSource):
    """
    ExtractedSource is a ControlSource whose messages are passed through an
    Extractor.
    """

    KEYS_BY_TYPE = {}
    NORMALIZERS = {}

    def __init__(self, callback, **kwds):
        """
        Arguments
        kwds -- keyword options passed to the  constructor of Extractor.
        """
        super().__init__(callback)
        self.extractor = extractor.Extractor(
            normalizers=self.NORMALIZERS,
            keys_by_type=self.KEYS_BY_TYPE,
            **kwds)

    def callback(self, msg):
        for m in self.extractor.extract(msg):
            self._callback(m)



PROGRAM = """\

WHAT IS THIS PROGRAM?

This is the frontend to BiblioPixel, the LED control system.

You can use this to run demos or your own projects, to set and get defaults
for projects, or to discover hardware devices attached to your computer.


PROJECTS

BiblioPixel projects are in JSON.  Each project is an object with four named
values called "sections".

  * driver: identifies the hardware driving the LED and its characteristics.
  * layout: represents the geometric layout of the LEDs in one or more drivers.
  * animation: a program that changes the LEDs over time.
  * run: time settings for the animation.

Any of these sections can be omitted, in which case BiblioPixel uses the
default values.

For more information, see
  https://github.com/ManiacalLabs/BiblioPixel/blob/master/README.md

"""
# TODO: finalize URL above.

LOGLEVEL = """\
"""



import copy

try:
    import numpy
    numpy_array = numpy.ndarray

except:
    numpy = None
    numpy_array = ()


class ListMath:
    @staticmethod
    def clear(color_list):
        if color_list:
            black = tuple(0 for i in color_list[0])
            color_list[:] = (black for i in color_list)

    @staticmethod
    def add(color_list, source, level, buffer):
        def add(color, src):
            return tuple(int(c + level * s) for (c, s) in zip(color, src))

        if level:
            color_list[:] = (add(c, s) for (c, s) in zip(color_list, source))

        return buffer

    @staticmethod
    def copy(color_list, source):
        color_list[:] = source


class NumpyMath:
    @staticmethod
    def clear(color_list):
        color_list.fill(0)

    @staticmethod
    def add(color_list, source, level, buffer):
        if level:
            buffer = numpy.multiply(source, level, out=buffer, casting='unsafe')
            color_list += buffer
        return buffer

    @staticmethod
    def copy(color_list, source):
        numpy.copyto(color_list, source, casting='unsafe')


def Math(color_list):
    return NumpyMath if isinstance(color_list, numpy_array) else ListMath


class ColorList:
    def __init__(self, color_list):
        self.color_list = color_list
        self.math = Math(color_list)
        self.buffer = None

    def clear(self):
        self.math.clear(self.color_list)

    def add(self, source, level):
        self.buffer = self.math.add(self.color_list, source, level, self.buffer)

    def copy_from(self, source):
        self.math.copy(self.color_list, source)

    def __len__(self):
        return len(self.color_list)

    def __getitem__(self, i):
        return self.color_list[i]


class Mixer:
    def __init__(self, color_list, sources, levels=None):
        self.color_list = ColorList(color_list)
        self.sources = sources
        self.levels = levels or []
        self.clear = self.color_list.clear

    def mix(self, master=1):
        for source, level in zip(self.sources, self.levels):
            self.color_list.add(source, level * master)



from . circle import (calc_ring_pixel_count, calc_ring_steps, make_circle_coord_map,
                      make_circle_coord_map_positions)
from . cube import make_cube_coord_map, make_cube_coord_map_positions
from . matrix import make_matrix_coord_map, make_matrix_coord_map_multi, make_matrix_coord_map_positions
from . strip import make_strip_coord_map, make_strip_coord_map_multi, make_strip_coord_map_positions

# DEPRECATED
from . rotation import Rotation, rotate_and_flip



def flatten(master):
    """
    Flattens a multilevel dictionary into a single-level one so that::

        {'foo':
            {'bar':
               {
                   'a': 1,
                   'b': True,
                   'c': 'hello',
                },
            },
        }

    would become::

        {'foo.bar.a': 1,
         'foo.bar.b': True,
         'foo.bar.a': 1,
         }

    You can mix and match both input (hierarchical) and output (dotted) formats
    in the input without problems - and if you call flatten more than once, it
    has no effect.
    """
    result = {}

    def add(value, *keys):
        if keys in result:
            raise ValueError('Duplicate key %s' % keys)
        result[keys] = value

    def recurse(value, *keys):
        if isinstance(value, dict):
            for k, v in value.items():
                recurse(v, k, *keys)
        else:
            key = '.'.join(reversed(keys))
            if key in result:
                raise ValueError('Duplicate key %s' % str(keys))
            result[key] = value

    recurse(master)
    return result



import traceback
from .. project import aliases, construct, project
from . import animation
from .. util import log


class Collection(animation.BaseAnimation):
    FAIL_ON_EXCEPTION = False

    @staticmethod
    def pre_recursion(desc):
        def cleanup_animation(a):
            if callable(a) or isinstance(a, str) or 'animation' not in a:
                animation = a
                a = {}
            else:
                animation = a.pop('animation')
            animation = construct.to_type_constructor(
                animation, 'bibliopixel.animation')
            arun = a.pop('run', {})
            arun = animation['run'] = dict(arun, **animation.get('run', {}))
            drun = desc['run']

            # Children without fps or sleep_time get it from their parents.
            if not ('fps' in arun or 'sleep_time' in arun):
                if 'fps' in drun:
                    arun.update(fps=drun['fps'])
                elif 'sleep_time' in drun:
                    arun.update(sleep_time=drun['sleep_time'])

            if a:
                raise ValueError('Extra fields in animation: ' + ', '.join(a))
            return animation

        desc['animations'] = [cleanup_animation(a) for a in desc['animations']]
        return desc

    CHILDREN = 'animations',

    def __init__(self, layout, animations=None, **kwds):
        super().__init__(layout, **kwds)
        self.animations = animations or []
        self.index = 0
        self.internal_delay = 0  # never wait

    # Override to handle all the animations
    def cleanup(self, clean_layout=True):
        self.state = animation.STATE.canceled
        for a in self.animations:
            if a:
                a.cleanup()
        super().cleanup(clean_layout)

    def add_animation(self, anim, **kwds):
        # DEPRECATED.
        anim.set_runner(kwds)
        self.animations.append(anim)

    def pre_run(self):
        self.index = -1

    @property
    def current_animation(self):
        if 0 <= self.index < len(self.animations):
            return self.animations[self.index]


class Parallel(Collection):
    def __init__(self, *args, levels, **kwds):
        super()(*args, **kwds)

        # Give each animation a unique, mutable layout so they can
        # run independently.
        for a in self.animations:
            a.layout = a.layout.mutable_copy()

    def step(self, amt=1):
        for a in self.animations:
            a.step(amt)


class Wrapper(Collection):
    def __init__(self, *args, source, **kwds):
        super()(*args, animations=[source], **kwds)
        self.animation = self.animations[0]

    def step(self, amt=1):
        self.animation.step(amt)



from . circle import Circle
from . matrix import Matrix
from . cube import Cube
from . pov import POV
from . strip import Strip
from . geometry.matrix import make_matrix_coord_map
from . geometry.circle import make_circle_coord_map
from . geometry.cube import make_cube_coord_map

# These are DEPRECATED
from . circle import LEDCircle
from . cube import LEDCube
from . matrix import LEDMatrix
from . pov import LEDPOV
from . strip import LEDStrip
from . geometry.rotation import Rotation



from . animation import BaseAnimation
from .. layout import Circle


class BaseCircleAnim(BaseAnimation):

    def __init__(self, layout, **kwds):
        super().__init__(layout, **kwds)

        if not isinstance(layout, Circle):
            raise RuntimeError('Must use bibliopixel.layout.Circle with ' +
                               'Circle Animations!')

        self.rings = layout.rings
        self.ringCount = layout.ringCount
        self.lastRing = layout.lastRing
        self.ringSteps = layout.ringSteps



import contextlib, threading, time
from . runner import Runner
from .. util import log
from .. util.threads.animation_threading import AnimationThreading
from .. project import attributes, fields
from enum import IntEnum


class STATE(IntEnum):
    ready = 0
    running = 1
    complete = 2
    canceled = 3
    max_steps = 4
    timeout = 5


class BaseAnimation(object):
    free_run = False
    pre_recursion = fields.default_converter

    @classmethod
    def construct(cls, project, *, run=None, name=None, data=None, **desc):
        """
        Construct an animation, set the runner, and add in the two
        "reserved fields" `name` and `data`.
        """
        a = cls(project.layout, **desc)
        a.set_runner(run or {})
        a.name = name
        a.data = data
        return a

    def __init__(self, layout, *, preclear=True, **kwds):
        attributes.set_reserved(self, 'animation', **kwds)
        self.layout = layout
        self.internal_delay = None
        self.on_completion = None
        self.state = STATE.ready
        self.preclear = preclear

    @property
    def _led(self):
        """Many BiblioPixelAnimations use the "protected" variable _led."""
        return self.layout

    @_led.setter
    def _led(self, layout):
        self.layout = layout

    @property
    def color_list(self):
        return self.layout.color_list

    @color_list.setter
    def color_list(self, cl):
        self.layout.color_list[:] = cl

    @property
    def completed(self):
        """Many BiblioPixelAnimations use the old `completed` variable."""
        return self.state == STATE.complete

    @completed.setter
    def completed(self, state):
        if state:
            self.state = STATE.complete
        else:
            self.state = STATE.running

    def pre_run(self):
        pass

    @property
    def title(self):
        return self.__class__.__name__

    def step(self, amt=1):
        raise RuntimeError("Base class step() called. This shouldn't happen")

    def cleanup(self, clean_layout=True):
        # if current thread is animation thread this was called
        # by the context manager and thread is therefore already stopped
        if self.threading.thread != threading.current_thread():
            self.threading.stop_thread(wait=True)
        # Some cases we may not want to clear the screen
        # Like with the remote, it would flash between anims
        if clean_layout:
            self.layout.cleanup()

    def compute_state(self):
        if self.threading.stop_event.isSet():
            self.state = STATE.canceled

        elif self.runner.seconds:
            elapsed = time.time() - self.runner.run_start_time
            if elapsed >= self.runner.seconds:
                self.state = STATE.timeout

        elif self.runner.max_steps:
            if self.cur_step >= self.runner.max_steps:
                self.state = STATE.max_steps

        elif not self.runner.until_complete:
            if self.state == STATE.complete:
                # Ignore STATE.complete if until_complete is False
                self.state = STATE.running

    def check_delay(self):
        if self.free_run:
            self.sleep_time = None
        elif self.internal_delay:
            self.sleep_time = self.internal_delay
        else:
            self.sleep_time = self.runner.sleep_time
        self.layout.animation_sleep_time = self.sleep_time or 0

    def run_one_frame(self):
        timestamps = []

        def stamp():
            timestamps.append(time.time())

        self.check_delay()

        stamp()

        self.step(self.runner.amt)

        stamp()

        self.layout.frame_render_time = timestamps[1] - timestamps[0]
        self.layout.push_to_driver()

        stamp()

        _report_framerate(timestamps)

        self.cur_step += 1
        if self.state == STATE.complete and self.runner.max_cycles > 0:
            if self.cycle_count < self.runner.max_cycles - 1:
                self.cycle_count += 1
                self.state = STATE.running

        stamp()

        self.threading.wait(self.sleep_time, timestamps)

        self.compute_state()

    @contextlib.contextmanager
    def run_context(self, clean_layout=True):
        self.state = STATE.running
        self.runner.run_start_time = time.time()
        self.threading.stop_event.clear()
        self._step = 0
        self.cur_step = 0
        self.cycle_count = 0

        self.check_delay()

        self.preclear and self.layout.all_off()
        self.pre_run()
        try:
            yield
        finally:
            self.cleanup(clean_layout)

        self.on_completion and self.on_completion(self.state)

        self.state = STATE.ready

    def run_all_frames(self, clean_layout=True):
        with self.run_context(clean_layout):
            while self.state == STATE.running:
                self.run_one_frame()

    def set_runner(self, runner):
        if isinstance(runner, Runner):
            self.runner = runner
        else:
            self.runner = Runner(**(runner or {}))
        self.threading = AnimationThreading(self.runner, self.run_all_frames)

    def start(self):
        self.threading.start()

    def run(self, **kwds):
        # DEPRECATED
        self.set_runner(kwds)
        self.start()


def _report_framerate(timestamps):
    total_time = timestamps[-1] - timestamps[0]
    fps = int(1.0 / max(total_time, 0.001))
    log.frame("%dms/%dfps / Frame: %dms / Update: %dms",
              1000 * total_time,
              fps,
              1000 * (timestamps[1] - timestamps[0]),
              1000 * (timestamps[2] - timestamps[1]))



from distutils.version import LooseVersion
from . codes import CMDTYPE, LEDTYPE, SPIChipsets, BufferChipsets
from ... drivers.return_codes import RETURN_CODES, raise_error
from ... util import log, util
from ... project.importer import import_module


class Devices(object):
    """Manage a list of serial devices."""

    def __init__(self, hardware_id, baudrate):
        self.hardware_id = hardware_id
        self.baudrate = baudrate
        self.serial = import_module('serial')
        self.Serial = self.serial.Serial
        self.list_ports = import_module('serial.tools.list_ports')

    def find_serial_devices(self):
        self.devices = {}
        hardware_id = "(?i)" + self.hardware_id  # forces case insensitive

        for ports in self.list_ports.grep(hardware_id):
            port = ports[0]
            try:
                id = self.get_device_id(port, self.baudrate)
                ver = self._get_device_version(port, self.baudrate)
            except:
                log.debug('Error getting device_id for %s, %s',
                          port, self.baudrate)
                continue

            if getattr(ports, '__len__', lambda: 0)():
                log.debug('Multi-port device %s:%s:%s with %s ports found',
                          self.hardware_id, id, ver, len(ports))
            if id < 0:
                log.debug('Serial device %s:%s:%s with id %s < 0',
                          self.hardware_id, id, ver)
            else:
                self.devices[id] = port, ver

        return self.devices

    def get_device(self, id=None):
        if id is None:
            if not self.devices:
                raise ValueError('No default device for %s' % self.hardware_id)
            id, (device, version) = sorted(self.devices.items())[0]

        elif id in self.devices:
            device, version = self.devices[id]

        else:
            error = 'Unable to find device with ID %s' % id
            log.error(error)
            raise ValueError(error)

        log.info("Using COM Port: %s, Device ID: %s, Device Ver: %s",
                 device, id, version)
        return id, device, version

    def error(self):
        error = "There was an unknown error communicating with the device."
        log.error(error)
        raise IOError(error)

    def set_device_id(self, dev, id, baudrate=921600):
        if id < 0 or id > 255:
            raise ValueError("ID must be an unsigned byte!")

        com = self.Serial(dev, baudrate=baudrate, timeout=5)

        packet = util.generate_header(CMDTYPE.SETID, 1)
        packet.append(id)
        com.write(packet)

        resp = com.read(1)
        if len(resp) == 0:
            self.error()
        elif ord(resp) != RETURN_CODES.SUCCESS:
            raise_error(ord(resp))

    def get_device_id(self, dev, baudrate=921600):
        packet = util.generate_header(CMDTYPE.GETID, 0)
        com = self.Serial(dev, baudrate=baudrate, timeout=5)
        com.write(packet)
        resp = ord(com.read(1))
        return resp

    def _get_device_version(self, dev, baudrate=921600):
        packet = util.generate_header(CMDTYPE.GETVER, 0)
        com = self.Serial(dev, baudrate=baudrate, timeout=0.5)
        com.write(packet)
        ver = 0
        resp = com.read(1)
        if len(resp) > 0:
            resp = ord(resp)
            if resp == RETURN_CODES.SUCCESS:
                ver = ord(com.read(1))
        return ver



import operator, os
from .. util import datafile, log

PROJECT_ALIASES = {}


BUILTIN_ALIASES = {
    # drivers
    'apa102': 'bibliopixel.drivers.SPI.APA102.APA102',
    'sk9822': 'bibliopixel.drivers.SPI.APA102.APA102',
    'dummy': 'bibliopixel.drivers.dummy_driver.Dummy',
    'hue': 'bibliopixel.drivers.hue.Hue',
    'lpd8806': 'bibliopixel.drivers.SPI.LPD8806.LPD8806',
    'mirror': 'bibliopixel.drivers.mirror.MirrorDriver',
    'network': 'bibliopixel.drivers.network.Network',
    'network_udp': 'bibliopixel.drivers.network_udp.NetworkUDP',
    'serial': 'bibliopixel.drivers.serial.Serial',
    'simpixel': 'bibliopixel.drivers.SimPixel.SimPixel',
    'ws281x': 'bibliopixel.drivers.SPI.WS281X.WS281X',
    'ws2801': 'bibliopixel.drivers.SPI.WS2801.WS2801',
    'spi': 'bibliopixel.drivers.SPI.SPI',
    'pi_ws281x': 'bibliopixel.drivers.PiWS281X.PiWS281X',

    # layouts
    'circle': 'bibliopixel.layout.circle.Circle',
    'cube': 'bibliopixel.layout.cube.Cube',
    'matrix': 'bibliopixel.layout.matrix.Matrix',
    'pov': 'bibliopixel.layout.pov.POV',
    'strip': 'bibliopixel.layout.strip.Strip',

    # animations
    'off': 'bibliopixel.animation.off.OffAnim',
    'feedback': 'bibliopixel.animation.feedback.Feedback',
    'matrix_calibration': 'bibliopixel.animation.tests.MatrixCalibrationTest',
    'matrix_test': 'bibliopixel.animation.tests.MatrixChannelTest',
    'mixer': 'bibliopixel.animation.mixer.Mixer',
    'receiver': 'bibliopixel.animation.receiver.BaseReceiver',
    'reprocess': 'bibliopixel.animation.reprocess.reprocess.Reprocess',
    'remote': 'bibliopixel.animation.remote.control.RemoteControl',
    'sequence': 'bibliopixel.animation.Sequence',
    'strip_test': 'bibliopixel.animation.tests.StripChannelTest',
}


def get_alias(alias):
    return PROJECT_ALIASES.get(alias) or BUILTIN_ALIASES.get(alias)
